import functools
import jsonschema
from falcon import HTTPBadRequest
from jsonschema import ValidationError
from inspect import getargspec


class ArgTypeError(TypeError):
    '''
    Use to raise exception when Argument type is invalid

    :param string func_name: Name of function
    :param int arg_oder: Order of arugment of function, base on zero
    :param object arg_value: Value of argumen
    :param string req_arg_type: Name of argument type is requiredt
    '''

    def __init__(self, func_name, arg_oder, arg_value, req_arg_type):
        self.error = 'The argument {} of {}() is not a {}. It is {}'.format(
            arg_oder, func_name, req_arg_type, type(arg_value)
        )

    def __str__(self):
        return self.error


class ArgNumberError(TypeError):
    '''
    Use to raise exception when number required arguments is not equal
    with number of function arguments

    :param string fn_name: Name of fucntion
    :param int num_args: Number of arguments in function
    '''

    def __init__(self, fn_name, num_args):
        self.error = 'Required full spec for {} arguments in {}()'.format(
            num_args, fn_name
        )

    def __str__(self):
        return self.error


class CtlArgError(TypeError):
    '''
    Use to raise exception when convert type in pre-controller processing

    :param object arg_value: Data to convert
    :param type req_arg: Required type
    '''

    def __init__(self, arg_value, req_arg):
        self.error = 'Can not convert <{}> to {} type'.format(
            arg_value, req_arg
        )

    def __str__(self):
        return self.error


def validate_args(fn_name, fn_args, req_args):
    '''
    Validate arguments of function. If type of element in req_args is dict,
    supported to validate by jsonschema. If type of element in req_args is,
    type, supported to compare type

    :param string fn_name: Name of function to be check
    :param tuple fn_args: Instance of arguments in order
    :param tuple req_args: Required type of arguments in Order

    :raise gwisp.validator.ArgNumberError: On number of required arguments
        is not enough
    :raise gwisp.validator.ArgTypeError: On type of arguments in function
        is invalid
    :raise jsonschema.exceptions.ValidationError: On value of arguments
        in function invalid jsonschema
    '''

    args = enumerate(zip(fn_args, req_args))

    for arg_num, (fn_arg, req_arg) in args:
        req_arg_type = type(req_arg)

        # check function type vs required type
        if req_arg_type is type:
            if type(fn_arg) is not req_arg:
                raise ArgTypeError(fn_name, arg_num, fn_arg, req_arg)

        # check function value with jsonschema
        elif req_arg_type is dict:
            jsonschema.validate(fn_arg, req_arg)

        # do not supported other case
        else:
            raise TypeError('Does not supported type of arguments')


def method(*req_args):
    '''
    This is decorator. Use to be check type of arguments in methods

    :param req_args: List of required arguments

    :raise gwisp.validator.ArgNumberError: On number of required arguments
        is not enough
    :raise gwisp.validator.ArgTypeError: On type of arguments in function
        is invalid
    :raise jsonschema.exceptions.ValidationError: On value of arguments
        in function invalid jsonschema
    '''

    def accept_decorator(fn):
        @functools.wraps(fn)
        def decorator_wrapper(*fn_args):
            argspec = getargspec(fn)
            fn_args_full = fn_args
            if argspec.defaults is not None:
                default_num = len(argspec.args) - len(fn_args)
                if default_num > 0:
                    fn_args_full = fn_args + argspec.defaults[-default_num:]

            if len(req_args) is not (len(fn_args_full) - 1):
                raise ArgNumberError(fn.__name__, len(fn_args_full))

            validate_args(fn.__name__, fn_args_full[1:], req_args)

            return fn(*fn_args)

        return decorator_wrapper

    return accept_decorator


def validate_req_msg(req_model, req):
    '''
    Validate HTTP message request of falcon.
    If req_model is None, do nothing. If req_model is not None, and it
    does not contains 'header' or 'body' key, TypeError will be raise.

    :param dict req_model: Data model of request in jsonschema. It include
        two key: header and body
    :param falcon.Request req: Request object
    :raise HTTPBadRequest: On request is invalid
    :raise TypeError: On req_model is invalid
    '''

    if req_model is None:
        return

    if 'header' not in req_model and 'body' not in req_model:
        raise TypeError('Request model is invalid')

    if 'body' not in req.context:
        raise HTTPBadRequest('400 Bad Request', 'Body is empty')

    try:
        if 'header' in req_model:
            jsonschema.validate(req.headers, req_model['header'])
        if 'body' in req_model:
            jsonschema.validate(req.context['body'], req_model['body'])
    except ValidationError as e:
        raise HTTPBadRequest(str(e))


def ctl(req_model, *req_args):
    '''
    This is decorator, use to validate request message and route url of
        falcon controller

    :param dict req_model: Data model of request in jsonschema. It include
        two key: header and body
    :param req_args: List of required type of arguments
    :raise HTTPBadRequest: On request is invalid
    :raise TypeError: On req_model is invalid
    '''

    def ctl_decorator(fn):
        @functools.wraps(fn)
        def decorator_wrapper(ctl_obj, req, res, **fn_args):
            # convert and validate route arguments
            route_args = getargspec(fn).args[3:]
            args = enumerate(zip(route_args, req_args))
            new_fn_args = []
            for i, (route_arg, req_arg) in args:
                try:
                    new_fn_args.append(req_arg(fn_args[route_arg]))
                except:
                    raise CtlArgError(fn_args[route_arg], req_arg)

            # validate request message
            validate_req_msg(req_model, req)

            # execute controller
            return fn(ctl_obj, req, res, *new_fn_args)

        return decorator_wrapper

    return ctl_decorator



import copy


from ..lowest import tracker
from ..repo import ctracker as rmodel_ctracker


res_body = copy.deepcopy(tracker.storage)
res_body['required'].extend(['_updated'])
res_body['properties'].update({
    '_updated': rmodel_ctracker.attr_updated
})
res = {'body': res_body}



from falcon import HTTP_204, HTTP_404, \
                   HTTPNotFound, HTTPForbidden
from bson.objectid import ObjectId

from ..error import NotExistError
from ..util import parse_selector
from ..validator import ctl
from ..repo import TrackerRepo
from ..model.front.post_tracker import req as fmodel_post
from ..model.front.put_tracker_id import req as fmodel_put


class TrackerCtl(object):
    '''
    Tracker resource

    :param pymongo.MongoClient db: Instance of mongo client
    :param gwisp.Protector protector: Instance of protector
    '''

    def __init__(self, db, protector):
        self._protector = protector
        self._repo = TrackerRepo(db)

    def on_get(self, req, res):
        '''
        Search trackers

        Request: Query parameters as selector

        Response: List of trackers match with selector
        '''

        res.body = self._repo.search(parse_selector(req))

        if res.body is None:
            res.status = HTTP_404

    @ctl(fmodel_post)
    def on_post(self, req, res):
        '''
        Create new tracker

        Request: Tracker in body, access token in header

        Response: Relative location of new Tracker in header, field
            Location
        '''

        auth = self._protector.check(req, iacc=True)
        item = req.context['body']

        # padding data
        item['author'] = {
            '_id': auth.account_id,
            'name': auth.account['name']
        }

        # save tracker to storage
        inserted_id = self._repo.insert_one(item)

        res.status = HTTP_204
        res.set_header('Location', '/tracker/{}'.format(inserted_id))


class TrackerItemCtl(object):
    '''
    Tracker resource

    :param pymongo.MongoClient db: Instance of mongo client
    :param gwisp.Protector protector: Instance of protector
    '''

    def __init__(self, db, protector):
        self._repo = TrackerRepo(db)
        self._protector = protector

    @ctl(None, ObjectId)
    def on_get(self, req, res, id):
        '''
        Query single tracker

        Request: Identify of tracker in request url

        Response: Tracker in body
        '''

        res.body = self._repo.find_by_id(id)

        if res.body is None:
            raise HTTPNotFound(
                title='404 Not Found',
                description='Not found tracker'
            )

    @ctl(fmodel_put, ObjectId)
    def on_put(self, req, res, id):
        '''
        Update tracker

        Request:
            - Identify of tracker in request url
            - Tracker content in body
            - Authorization header

        Response: No Content
        '''

        auth = self._protector.check(req)
        old_item = self._repo.find_by_id(id)
        if old_item is None:
            raise NotExistError('Tracker is not exist')
        if old_item['author']['_id'] != auth.account_id:
            raise HTTPForbidden(
                title='403 Forbidden',
                description='Can not update item of other account'
            )

        # put data to storage
        self._repo.update_one(id, req.context['body'])

        # result is no content
        res.status = HTTP_204

    @ctl(None, ObjectId)
    def on_delete(self, req, res, id):
        '''
        Remove tracker

        Request: Identify of tracker in request url
        '''

        # authenticate
        auth = self._protector.check(req)

        # ensure item is exist
        item = self._repo.find_by_id(id)
        if item is None:
            raise HTTPNotFound(description='Tracker is not exist')

        # ensure only owner have permission to delete item
        if item['author']['_id'] != auth.account_id:
            raise HTTPForbidden(
                title='403 Forbidden',
                description='Can not delete item of other account'
            )

        # remove item from storage
        self._repo.delete_one(id)

        # result is no content
        res.status = HTTP_204



from ..lowest import account as lmodel_account


req_body = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'language': lmodel_account.attr_language,
        'avatar': lmodel_account.attr_avatar
    }
}
req = {'body': req_body}



from falcon import HTTPNotFound, HTTP_204
from bson import ObjectId

from ..validator import ctl
from ..util import parse_selector
from ..repo import CTrackerRepo


class TrackerClonedCtl(object):
    def __init__(self, db, protector):
        self._ctracker_repo = CTrackerRepo(db)
        self._protector = protector

    def on_get(self, req, res):
        selector = parse_selector(req)
        auth = self._protector.check(req)

        res.body = self._ctracker_repo.search_on_cview(
            auth.account_id, selector
        )

        if res.body is None:
            raise HTTPNotFound(description='Not found your trackers')

    def on_put(self, req, res):
        auth = self._protector.check(req)

        self._ctracker_repo.update_all_cloned(auth.account_id)

        res.status = HTTP_204


class TrackerClonedItemCtl(object):
    def __init__(self, db, protector):
        self._ctracker_repo = CTrackerRepo(db)
        self._protector = protector

    @ctl(None, ObjectId)
    def on_get(self, req, res, id):
        auth = self._protector.check(req)

        res.body = self._ctracker_repo.find_by_id_on_cview(auth.account_id, id)

        if res.body is None:
            raise HTTPNotFound(description='Not found cloned trackers')

    @ctl(None, ObjectId)
    def on_put(self, req, res, id):
        auth = self._protector.check(req)

        self._ctracker_repo.update_one_cloned(auth.account_id, id)

        res.status = HTTP_204

    @ctl(None, ObjectId)
    def on_post(self, req, res, id):
        auth = self._protector.check(req)

        self._ctracker_repo.clone_one(auth.account_id, id)

        res.status = HTTP_204
        res.append_header(
            'Location', '/tracker/cloned/{}'.format(id)
        )

    @ctl(None, ObjectId)
    def on_delete(self, req, res, id):
        auth = self._protector.check(req)

        self._ctracker_repo.remove_one_cloned(auth.account_id, id)

        res.status = HTTP_204



import re
from math import ceil
from time import time
from bson import ObjectId

from .tracker import TrackerRepo

from ..error import NotExistError
from ..util import Selector
from ..validator import method
from ..model.engine import ctracker as emodel_ctracker


search_by_author_fields = {
    'tags': 0, 'tasks': 0, 'notes': 0
}
search_by_cloner_fields = {
    'tags': 0, 'tasks': 0, 'notes': 0, '_cloner': 0
}


class CTrackerRepo(object):
    '''
    Provide methods to manage cloned trackers. It will help user use
    trackers without affected from updating by author of trackers

    :param pymongo.MongoClient db: Instance of connection to mongdb server
    '''
    def __init__(self, db):
        self._tracker_repo = TrackerRepo(db)
        self._tracker_coll = db['tracker']
        self._ctracker_coll = db['ctracker']
        self._acc_coll = db['account']

    @method(ObjectId, ObjectId)
    def clone_one(self, cloner, tracker_id):
        '''
        Create an cloned tracker belong account if it is not early cloned.
        Replace data of tracker if it is early cloned

        :param bson.objectid.ObjectId cloner: Account identity
            which is clone tracker
        :param bson.objectid.ObjectId tracker_id: Tracker identity to clone
        '''

        # Find tracker to clone
        real_tracker = self._tracker_repo.find_by_id(tracker_id)
        if real_tracker is None:
            raise NotExistError('Tracker is not exist')

        # padding data
        real_tracker['_cloner'] = cloner
        real_tracker['_cloned_date'] = int(time())

        # validate data
        emodel_ctracker.validate_insert(real_tracker)

        # insert or update tracker depend on it is early exists or not
        if (self._is_exist(cloner, tracker_id)):
            self._ctracker_coll.update_one(
                {'_id': tracker_id, '_cloner': cloner},
                {'$set': real_tracker}
            )
        else:
            self._ctracker_coll.insert_one(real_tracker)

    @method(ObjectId, Selector)
    def search_on_uview(self, observer, selector=Selector()):
        '''
        Search trackers on user view correspond with observer.
        On that view, observer see that trackers is cloned by they or not.
        And if trackers is cloned by they, it is updated with real trackers
        or not.
        Two field are added to each trackers: **_cloned** and **_updated**

        :param bson.objectid.ObjectId observer: Identity of account
        :param gwisp.Selector selector: Keyword and pagging information
        :return: List of tracker match with selector. None on not found
        :rtype: list
        :rtype: None
        '''

        # search real trackers
        real_items = self._tracker_repo.search(selector)
        if real_items is None:
            return None

        # search cloned trackers of account
        real_ids = [i['_id'] for i in real_items]
        cloned_cusor = self._ctracker_coll.find({
            '_cloner': observer,
            '_id': {'$in': real_ids}
        })
        cloned_items = list(cloned_cusor)

        # compare real and cloned trackers to mark '_cloned' and '_updated'
        cloned_ids = [i['_id'] for i in cloned_items]
        cloned_index = {i['_id']: i for i in cloned_items}
        for item in real_items:
            # tracker is not cloned
            if item['_id'] not in cloned_ids:
                item['_cloned'] = False
                item['_updated'] = False
                continue

            # tracker is cloned
            item['_cloned'] = True
            cloned_item = cloned_index[item['_id']]
            if item['modified'] > cloned_item['_cloned_date']:
                item['_updated'] = False
            else:
                item['_updated'] = True

        return real_items

    @method(ObjectId, Selector)
    def search_on_cview(self, cloner, selector=Selector()):
        '''
        Search cloned trackers on cloner view itself. On that view,
        cloner see that cloned trackers is updated with real trackers
        or not. One field added: '_updated'

        :param bson.objectid.ObjectId cloner: Account identity
        :param gwisp.Selector selector: Search information
        :return: List of trackers was found. None on not found
        :rtype: list
        '''

        # search trackers is cloned by account
        filters = {'$and': [{'_cloner': cloner}]}
        if selector.keyword is not None:
            filters['$and'].append(self._gen_filter(selector.keyword))

        cloned_cusor = self._ctracker_coll.find(
            filters,
            search_by_cloner_fields,
            skip=selector.skip,
            limit=selector.page_size
        )
        cloned_items = list(cloned_cusor)
        if len(cloned_items) == 0:
            return None

        # search real trackers correspond with cloned trackers
        cloned_ids = [i['_id'] for i in cloned_items]
        real_items = self._tracker_coll.find(
            {'_id': {'$in': cloned_ids}},
            {'notes': 0, 'tags': 0, 'tasks': 0, 'author': 0, 'num_user': 0}
        )

        # compare cloned trackers with real trackers to mar '_updated'
        real_indexs = {i['_id']: i['modified'] for i in real_items}
        for item in cloned_items:
            if real_indexs[item['_id']] > item['_cloned_date']:
                item['_updated'] = False
            else:
                item['_updated'] = True

            # remove data
            del item['_cloned_date']

        return cloned_items

    @method(ObjectId, ObjectId)
    def find_by_id_on_uview(self, observer, tracker_id):
        '''
        Find one tracker on user view correspond with observer.
        On that view, observer see that tracker is cloned by they or not.
        And if tracker is cloned by they, it is updated with real tracker
        or not. Two fields are added: **_cloned** and **_updated**

        :param bson.objectid.ObjectId observer: Account identity
        :param bson.objectid.ObjectId tracker_id: Tracker identity
        :return dict: Real tracker data on found. None on not found
        :rtype: dict
        :rtype: None
        '''

        # search real tracker
        real_item = self._tracker_repo.find_by_id(tracker_id)
        if real_item is None:
            return None

        # search cloned tracker
        cloned_item = self._ctracker_coll.find_one({
            '_id': tracker_id,
            '_cloner': observer
        })

        # tracker is not cloned
        if cloned_item is None:
            real_item['_cloned'] = False
            real_item['_updated'] = False
            return real_item

        # tracker is cloned, mark '_updated' field
        real_item['_cloned'] = True
        if real_item['modified'] > cloned_item['_cloned_date']:
            real_item['_updated'] = False
        else:
            real_item['_updated'] = True

        return real_item

    @method(ObjectId, ObjectId)
    def find_by_id_on_cview(self, cloner, tracker_id):
        '''
        Find cloned tracker on cloner view itself. On that view, cloner see
        that cloned tracker is updated with real tracker or not.
        One field added: ‘_updated’

        :param bson.objectid.ObjectId cloner: Cloner identity
        :param bson.objectid.ObjectId tracker_id: Tracker identity
        :return: Cloned tracker data on found. None on not found
        :rtype: dict
        :rtype: None
        '''

        # find real tracker
        real_item = self._tracker_repo.find_by_id(tracker_id)
        if real_item is None:
            return None

        # find cloned tracker
        cloned_item = self._ctracker_coll.find_one(
            {'_id': tracker_id, '_cloner': cloner}
        )
        if cloned_item is None:
            return None

        # compare real and cloned tracker to mark '_updated' field
        if real_item['modified'] > cloned_item['_cloned_date']:
            cloned_item['_updated'] = False
        else:
            cloned_item['_updated'] = True

        # remove data
        del cloned_item['_cloner']
        del cloned_item['_cloned_date']

        return cloned_item

    @method(ObjectId, ObjectId)
    def use_tracker(self, account_id, tracker_id):
        '''
        Make account with account_id use tracker with tracker_id.
        If tracker is not cloned, clone it.
        Then copy cloned version to account.tracker

        :param bson.objectid.ObjectId account_id: Account identity
        :param bson.objectid.ObjectId tracker_id: Tracker identity
        '''

        # find cloned tracker
        cloned_item = self._ctracker_coll.find_one({
            '_cloner': account_id,
            '_id': tracker_id
        })

        # if tracker is not cloned, clone it
        if cloned_item is None:
            cloned_item = self._tracker_repo.find_by_id(tracker_id)
            if cloned_item is None:
                raise NotExistError('Tracker is not exists')

            cloned_item['_cloner'] = account_id
            cloned_item['_cloned_date'] = int(time())

            self._ctracker_coll.insert_one(cloned_item)

        # copy cloned tracker to account.tracker
        del cloned_item['_cloner']
        self._acc_coll.update_one(
            {'_id': account_id},
            {'$set': {'tracker': cloned_item}}
        )

    @method(ObjectId)
    def update_all_cloned(self, cloner):
        '''
        Update all of cloned trackers is cloned by cloner with real
        trackers

        :param bson.objectid.ObjectId cloner: Account identity
        '''

        # find trackers is using by account
        account = self._acc_coll.find_one(
            {'_id': cloner},
            {'email': 0, 'avatar': 0, 'groups': 0, 'subject': 0}
        )
        if account is None:
            raise NotExistError('Account is not exist')

        # prepare for cloned trackers
        number_item = self._ctracker_coll.count({'_cloner': cloner})
        page_size = 16
        number_of_page = ceil(number_item / page_size)

        # clone each page of trackers
        for page_index in range(number_of_page):
            # query page of cloned trackers
            cloned_cusor = self._ctracker_coll.find(
                {'_cloner': cloner},
                {
                    'tags': 0, 'tasks': 0, 'notes': 0, 'author': 0,
                    'modified': 0
                },
                skip=page_index*page_size,
                limit=page_size
            )
            cloned_items = list(cloned_cusor)

            # query page of real trackers
            cloned_ids = [i['_id'] for i in cloned_items]
            real_cusor = self._tracker_coll.find(
                {'_id': {'$in': cloned_ids}}
            )
            real_items = list(real_cusor)

            # find trackers is using by account then update
            # and padding meta data for each real trackers
            for item in real_items:
                item['_cloned_date'] = int(time())

                if 'tracker' in account:
                    if item['_id'] == account['tracker']['_id']:
                        self._acc_coll.update_one(
                            {'_id': cloner},
                            {'$set': {'tracker': item}}
                        )

                item['_cloner'] = cloner

            # remove all of cloned trackers
            self._ctracker_coll.remove({'_id': {'$in': cloned_ids}})

            # create new cloned trackers
            self._ctracker_coll.insert(real_items)

    @method(ObjectId, ObjectId)
    def update_one_cloned(self, cloner, tracker_id):
        '''
        Update one cloned tracker.
        If tracker is using by cloner, update it

        :param bson.objectid.ObjectId cloner: Account identity
        :param bson.objectid.ObjectId tracker_id: Tracker identity
        '''

        # find real tracker
        real_item = self._tracker_coll.find_one({'_id': tracker_id})
        if real_item is None:
            raise NotExistError('Not found tracker')

        # find account
        account = self._acc_coll.find_one({'_id': cloner})
        if account is None:
            raise NotExistError('Account is not exist')

        # find cloned tracker
        cloned_item = self._ctracker_coll.find_one({
            '_id': tracker_id,
            '_cloner': cloner
        })
        if cloned_item is None:
            raise NotExistError('Tracker is not cloned')

        # padding meta data to tracker
        real_item['_cloned_date'] = int(time())

        # update using tracker if match
        if 'tracker' in account and account['tracker']['_id'] == tracker_id:
                self._acc_coll.update_one(
                    {'_id': cloner},
                    {'$set': {'tracker': real_item}}
                )

        # update cloned tracker
        self._ctracker_coll.update_one(
            {'_id': tracker_id, '_cloner': cloner},
            {'$set': real_item}
        )

    @method(ObjectId, ObjectId)
    def remove_one_cloned(self, cloner, tracker_id):
        '''
        Remove cloned tracker from account.
        If cloned tracker is using by account, remove it

        :param bson.objectid.ObjectId cloner: Account identity
        :param bson.objectid.ObjectId tracker_id: Tracker identity
        '''

        # remove cloned tracker
        result = self._ctracker_coll.delete_one({
            '_id': tracker_id,
            '_cloner': cloner
        })
        if result.deleted_count == 0:
            raise NotExistError('Tracker is not exist')

        # remove using tracker if match
        account = self._acc_coll.find_one({'_id': cloner})
        if account is None:
            raise NotExistError('Account is not exist')
        if 'tracker' in account and account['tracker']['_id'] == tracker_id:
            self._acc_coll.update_one(
                {'_id': cloner},
                {'$unset': {'tracker': None}}
            )

    @method(ObjectId, ObjectId)
    def _is_exist(self, account_id, tracker_id):
        '''
        Check tracker is cloned by account

        :param bson.objectid.ObjectId account_id: Account identity
        :param bson.objectid.ObjectId tracker_id: Tracker identity
        :return: True on tracker is cloned by account
        :return: False on tracker is not cloned by account
        :rtype: bool
        '''

        key = {'account_id': account_id, '_id': tracker_id}
        return self._ctracker_coll.find_one(key) is not None

    def _gen_filter(self, keyword):
        '''
        Generate query filter of mongodb

        :param str keyword: Keyword to search
        :return: Filter
        :rtype: dict
        '''
        return {'$or': [
            {'name': {'$regex': re.compile(keyword)}},
            {'tags': {
                '$elemMatch': {
                    '$regex': re.compile(keyword)
                }
            }}
        ]}



from falcon import HTTPNotFound, HTTPInternalServerError, HTTP_204
from bson import ObjectId

from ..validator import ctl
from ..repo import TrackerRepo, CTrackerRepo


class MeTrackerCtl(object):
    def __init__(self, db, protector):
        self._tracker_repo = TrackerRepo(db)
        self._protector = protector

    def on_get(self, req, res):
        auth = self._protector.check(req, iacc=True)

        if 'tracker' not in auth.account:
            raise HTTPNotFound(description='No tracker is using')
        target = auth.account['tracker']

        real_item = self._tracker_repo.find_by_id(target['_id'])
        if real_item is None:
            raise HTTPInternalServerError(
                description='Not found real tracker'
            )

        # mark tracker is updated with real tracker or not
        if real_item['modified'] > target['_cloned_date']:
            target['_updated'] = False
        else:
            target['_updated'] = True

        # remove data
        del target['_cloned_date']

        res.body = target


class MeTrackerItemCtl(object):
    def __init__(self, db, protector):
        self._protector = protector
        self._ctracker_repo = CTrackerRepo(db)

    @ctl(None, ObjectId)
    def on_put(self, req, res, id):
        auth = self._protector.check(req)

        self._ctracker_repo.use_tracker(auth.account_id, id)

        res.status = HTTP_204



from .root import RootCtl
from .oauth import OauthCtl, LoginUrlCtl, LoginCtl, MeCtl
from .tracker import TrackerCtl, TrackerItemCtl
from .tracker_account import TrackerAccountCtl, TrackerAccountItemCtl
from .tracker_me import TrackerMeCtl
from .me_tracker import MeTrackerCtl, MeTrackerItemCtl
from .tracker_cloned import TrackerClonedCtl, TrackerClonedItemCtl



from falcon import HTTPNotFound
from bson import ObjectId

from ..util import parse_selector
from ..validator import ctl
from ..repo import CTrackerRepo


class TrackerAccountCtl(object):
    '''
    Allow search trackers on user view

    :param pymongo.MongoClient db: Instance of connection to mongodb server
    :param gwisp.Protector protector: Instance of protector
    '''

    def __init__(self, db, protector):
        self._protector = protector
        self._repo = CTrackerRepo(db)

    def on_get(self, req, res):
        auth = self._protector.check(req)
        selector = parse_selector(req)

        res.body = self._repo.search_on_uview(auth.account_id, selector)
        if res.body is None:
            raise HTTPNotFound(description='Not found trackers')


class TrackerAccountItemCtl(object):
    '''
    Allow query single shceduler on user view point

    :param pymongo.MongoClient db: Instance of connection to mongodb server
    :param gwisp.Protector protector: Instance of protector
    '''

    def __init__(self, db, protector):
        self._protector = protector
        self._repo = CTrackerRepo(db)

    @ctl(None, ObjectId)
    def on_get(self, req, res, id):
        auth = self._protector.check(req)

        res.body = self._repo.find_by_id_on_uview(auth.account_id, id)
        if res.body is None:
            raise HTTPNotFound(
                title='404 Not Found',
                description='Not found tracker'
            )



import copy

from .get_tracker import res_body as get_tracker_res_body
from ..repo import ctracker as rmodel_ctracker

res_body = copy.deepcopy(get_tracker_res_body)
res_body['items']['required'].extend(['_updated'])
res_body['items']['properties'].update({
    '_updated': rmodel_ctracker.attr_updated
})
res = {'body': res_body}
